/*!
 * © 2022 David Vines
 *
 * domainOfTheAncients is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * any later version.
 *
 * domainOfTheAncients is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with domainOfTheAncients. If not, see https://www.gnu.org/licenses/gpl-2.0.html.
 */
/* globals ROT, Map, Player, Util, PlayerOrders, SensorOrders, ShipOrderReport, ShipDesignType, ShipDesign, ShipDesigner, JumpShipOrder, JumpToAndColonisePlanetShipOrder, HoldPositionShipOrder, HoldAndColoniseShipOrder */
/* exported stylenames */

const InitialState = { version: "25", turn: 1 }; /* Initial Game State - version indicates which version of save data we create and can read */
const stylenames = ['player', 'alkari', 'auron', 'caffe', 'clanger', 'drahmini', 'gobb', 'iyanden', 'jerandi', 'kalli', 'kelle',
	'pican', 'regul', 'saggitt', 'strata', 'fleet', 'worati'];
const techs = [0, 8, 9, 6, 6, 5, 3, 1, 8, 8, 3, 9, 4, 5, 3, 2, 0];
var Game = {
	STYLENAMES: stylenames,
	FULLNAMES: {'player': "you",
				'alkari': "the Alkari",
				'auron':"the Aurons",
				'caffe':"the Caffenians",
				'clanger':"the Clangers",
				'drahmini':"the Drahmini",
				'gobb':"the Gobbannanean Principalities",
				'iyanden':"the Chroniomancy of Iyanden",
				'jerandi':"the Jerandi People's Republic'",
				'kalli':"the Kallikanzaroi",
				'kelle':"the Kellesonians",
				'pican':"the Picans",
				'regul':"the Regul",
				'saggitt':"the Saggittarians",
				'strata':"Strata Inc.",
				'fleet':"the Fleet",
				'worati':"the Worati"
				},
	State: { version: undefined, baseMap: undefined, playerMap: undefined, turn: undefined, playerProduction: undefined },
	_standardShipDesigns: [],
	_players: [],
	History: { baseMap: {} },
	_showMap: function(map) {
		map.forEachHex(function(hex) {
			const hexelement = document.getElementById(hex.id);
			if (hex.lastScanned) {
				const planetElement = hexelement.childNodes[3];
				planetElement.classList.remove("emptyspace", "starbase","planet");
				let style = "emptyspace";
				if (hex.starbase) style = "starbase";
				if (hex.planet) style = "planet";
				planetElement.classList.add(style);
				hexelement.classList.remove(...stylenames);
				if (hex.lastScanned && hex.owner) {
					hexelement.classList.add(hex.owner);
				}
				let text = "";
				if (hex.level) {
					text = "Class "+hex.level+" planet\n";
				}
				if (hex.owner) {
					text += "Owned by "+Game.FULLNAMES[hex.owner]+"\n";
				}
				hexelement.title = text + "Last scanned on Turn " + hex.lastScanned;
				hexelement.classList.remove("fogofwar");
			} else {
				hexelement.classList.add("fogofwar");
			}
		});
	},
	_generateNewGame: function() {
		Game.State = Object.assign({}, InitialState);
		Game.State.baseMap = new Map();
		Game.History = { baseMap: {}, version: Game.State.version };
		Game.State.baseMap.forEachHex(function(hex) {
			if (hex.planet) Game.History.baseMap[hex.id] = true;
		}); /* basemap shows where the planets are */

		// Now create the players
		Game._createPlayers(document.getElementById("favouredTech").selectedIndex);

		document.getElementById("welcome").classList.replace("visible", "hidden");
		document.getElementById("summary").classList.replace("hidden", "visible");
		PlayerOrders.setPlayer(Game._players[0]);
		PlayerOrders.setStandardShipDesigns(Game._standardShipDesigns);
		Game.State.turn = 0;
		Game._newTurn();
	},
	_createPlayers: function(playerTechIndex) {
		const capital = {};
		Game.State.baseMap.forEachHex(hex => {
			if (hex.owner) {
				capital[hex.owner] = hex.id;
			}
		});
		Game._players = [];
		for(let i=0; i< Game.STYLENAMES.length; i++) {
			let player = new Player(Game.STYLENAMES[i],capital[Game.STYLENAMES[i]],(i === 0 ? playerTechIndex : techs[i]));
			player.init(Game.State.baseMap,Game.State.turn);
			Game._players.push(player);
		}
	},
	_newTurn: function() {
		Game.State.turn++;
		Game._updateHistory();
		Game._saveGameState();
		Game._startTurn();
	},
	_startTurn: function() {
		Game._clearDisplayTables();
		Game._showResearchReports();
		Game._showColonyReports();
		ShipOrderReport.showShipReports(Game._players[0]);
		SensorOrders.showReports(Game._players[0],Game.State.turn);
		Game._showMap(Game._players[0].map); // After the sensors have had a chance to scan hexes for the map
		PlayerOrders.prepare();
		Util.replaceText("turnnumber", Game.State.turn);
	},
	_clearDisplayTables: function() {
	// Clear existing rows after the templates
		for(let templaterow of document.querySelectorAll(".template")) {
			while (templaterow.nextSibling) {
				templaterow.parentNode.removeChild(templaterow.nextSibling);
			}
		}
	},
	// TODO Side effect: Also compute total production - this may (probably will) need to be refactored at some point
	_showColonyReports: function() {
		// Find the colony listing table
		let colonies = document.getElementById("colonies");
		let production = 0;
		Game._players[0].map.forEachHex(function(hex) {
			if (hex.owner == Game.STYLENAMES[0] && hex.level > 0) {
				const newrow = colonies.insertRow();
				const rowname = document.createElement("td"); rowname.append(hex.id);
				const report = document.createElement("td"); report.append("Nothing to report");
				const marines = document.createElement("td"); marines.append(hex.marines);
				const productionCell = document.createElement("td"); productionCell.append(hex.level * 20);
				newrow.append(rowname,report,marines,productionCell);
				production += hex.level * 20;
				Game._players[0].map.setLastScanned(hex.id,Game.State.turn);
			}
		});
		Game._players[0].production = production;
	},
	_showResearchReports: function() {
		Game._players[0].forEachTech( (tech, research) => {
			Util.replaceText(tech.domPrefix+"Points",research);
			let level = tech.level(research);
			Util.replaceText(tech.domPrefix+"Level",level);
			while(level > 1) {
				document.getElementById(tech.domPrefix+level).classList.remove("hidden");
				level--;
			}
		});
	},
	_updateHistory: function() {
		let sheet = {};
		Game.State.baseMap.forEachHex(function(hex) {
			if (hex.owner) sheet[hex.id] = hex.owner; /* Each turn has a map showing who owns the hex */
		});
		Game.History[Game.State.turn] = sheet;
		localStorage.domainOfTheAncientsHistory = JSON.stringify(Game.History);
	},
	init: function() {
		// Set up the standard ship types (the order will be reversed when the expenditure screen shows them (so colony ship will initially be at top of the list)
		const artifact = new ShipDesign("Artifact",ShipDesignType.Ship);
		artifact.addBlock("_b8"); artifact.addBlock("_p7"); artifact.addBlock("_p7"); artifact.addBlock("_jh"); artifact.addBlock("_mh");
		artifact.addBlock("_wx"); artifact.addBlock("_an"); artifact.addBlock("_dv"); artifact.addBlock("_dv"); artifact.addBlock("_dv");
		Game._standardShipDesigns.push(artifact);

		const exampleSDB = new ShipDesign("Standard System Defense Boat",ShipDesignType.Ship);
		exampleSDB.addBlock("_b1"); exampleSDB.addBlock("_p1"); exampleSDB.addBlock("_p1"); exampleSDB.addBlock("_p1");
		exampleSDB.addBlock("_ma"); exampleSDB.addBlock("_ma"); exampleSDB.addBlock("_ma");
		exampleSDB.addBlock("_wl"); exampleSDB.addBlock("_wl"); exampleSDB.addBlock("_wl");
		Game._standardShipDesigns.push(exampleSDB);

		const exampleFighter = new ShipDesign("Standard TL1 fighter",ShipDesignType.Ship);
		exampleFighter.addBlock("_b1"); exampleFighter.addBlock("_p1"); exampleFighter.addBlock("_p1"); exampleFighter.addBlock("_ja");
		exampleFighter.addBlock("_ma"); exampleFighter.addBlock("_ma"); exampleFighter.addBlock("_wl");
		exampleFighter.addBlock("_wl"); exampleFighter.addBlock("_wl"); exampleFighter.addBlock("_da");
		Game._standardShipDesigns.push(exampleFighter);

		const exampleColonyShip = new ShipDesign("Standard Colony Ship",ShipDesignType.Ship);
		exampleColonyShip.addBlock("_b0"); exampleColonyShip.addBlock("_p1"); exampleColonyShip.addBlock("_ja"); exampleColonyShip.addBlock("_ma");
		exampleColonyShip.addBlock("_cb"); exampleColonyShip.addBlock("_cb"); exampleColonyShip.addBlock("_cb");
		exampleColonyShip.addBlock("_cb"); exampleColonyShip.addBlock("_cb"); exampleColonyShip.addBlock("_cb");
		Game._standardShipDesigns.push(exampleColonyShip);

		const exampleSpaceBase = new ShipDesign("Standard Space Base",ShipDesignType.SpaceBase);
		exampleSpaceBase.addBlock("_b1"); exampleSpaceBase.addBlock("_p1"); exampleSpaceBase.addBlock("_wl"); exampleSpaceBase.addBlock("_wl");
		exampleSpaceBase.addBlock("_wl"); exampleSpaceBase.addBlock("_wl"); exampleSpaceBase.addBlock("_wl");
		exampleSpaceBase.addBlock("_wl"); exampleSpaceBase.addBlock("_wl"); exampleSpaceBase.addBlock("_wl");
		Game._standardShipDesigns.push(exampleSpaceBase);

		const exampleStarBase = new ShipDesign("Standard Star Base",ShipDesignType.StarBase);
		exampleStarBase.addBlock("_b1"); exampleStarBase.addBlock("_p1"); exampleStarBase.addBlock("_wl"); exampleStarBase.addBlock("_wl");
		exampleStarBase.addBlock("_wl"); exampleStarBase.addBlock("_wl"); exampleStarBase.addBlock("_wl");
		exampleStarBase.addBlock("_wl"); exampleStarBase.addBlock("_wl"); exampleStarBase.addBlock("_wl");
		Game._standardShipDesigns.push(exampleStarBase);

		Game._initGameButtons();
		Game._initTwisties();
		PlayerOrders.init();
		ShipDesigner.init();
	},
	_initGameButtons: function() {
		document.getElementById("generate").onclick = Util.captureExceptionFunction(Game._generateNewGame);
		if (localStorage.domainOfTheAncients && localStorage.domainOfTheAncientsVersion && localStorage.domainOfTheAncientsSaveDate) {
			Game._showLoadButton();
			let v;
			try {
				v = JSON.parse(localStorage.domainOfTheAncientsVersion);
			} catch(_e) {
				v = null; // Clearly no version data if we can't parse it!
			}
				if (v && v.game && (v.game == InitialState.version)) {
					document.getElementById("resume").disabled = false;
					document.getElementById("wrongVersion").classList.replace("visible","hidden");
					document.getElementById("resume").onclick = Util.captureExceptionFunction(Game._load);
				} else {
					document.getElementById("resume").disabled = true;
					if (v && v.plugin) {
						Util.replaceText("incompatibleVersionNumber",v.plugin);
					}
					document.getElementById("wrongVersion").classList.replace("hidden","visible");
				}
		} else {
			document.getElementById("resume").disabled = true;
		}
		if (localStorage.domainOfTheAncientsHistory) {
			// Nothing yet!
		} else {
			document.getElementById("replay").disabled = true;
		}

		document.getElementById("nextTurn").onclick = Util.captureExceptionFunction(Game._processTurn);
		document.getElementById("processorderread").onclick = Util.captureExceptionFunction(Game._completeTurn);
	},
	_initTwisties: function() {
		// Look for the twisties and set them up
		for(let button of document.querySelectorAll("h4 button")) {
			// Look for the child svg's 'we are interested in
			let svg1 = null, svg2 = null, followingdiv = null;
			for(let child of button.childNodes) {
				if (child.nodeName == "svg") {
					if (svg1 === null) {
						svg1 = child;
					} else {
						svg2 = child;
					}
				}
			}
			// And then the uncle's' (i.e. parent's sibling) div
			let sibling = button.parentNode.nextSibling;
			while(followingdiv === null && sibling) {
				if (sibling.nodeName === "DIV") {
					followingdiv = sibling;
				} else {
					sibling = sibling.nextSibling;
				}
			}

			if (svg1 && svg2 && followingdiv) {
				button.onclick = Util.captureExceptionFunction(Util.toggleFields,[svg1,svg2,followingdiv]);
			}
		}
	},
	_processTurn: function() {
		const summary = document.getElementById("summary");
		const proc = document.getElementById("processOrder");
		const playerOrders = PlayerOrders.getOrders();
		const updates = [];

		// Ship orders
		for(let player of Game._players) {
			let shipsToBeDestroyed = [];
			for(let ship of player.ships) {
				// TODO Refactor this code at some point!
				let report = null;
				if (ship.order instanceof JumpShipOrder || ship.order instanceof JumpToAndColonisePlanetShipOrder) {
					ship.location = ship.order.target;
					if (player === Game._players[0])  {// TODO the player map probably needs to move into the player (with a unique map per player)
						const wasUnscanned = (player.map.hex(ship.location).lastScanned ? false : true);
						const targetHex = player.map.hex(ship.location);
						player.map.setLastScanned(targetHex,Game.State.turn);
						if (targetHex.planet && targetHex.owner) {
							ship.report = `${ship.location} is a class ${targetHex.level} colony belonging to ${Game.FULLNAMES[targetHex.owner]}`;
						} else if (targetHex.planet) {
							ship.report = `${ship.location} is an uncolonised world`;
						}
						if (wasUnscanned && ship.hasAReport()) {
							report = ship.report;
						}
					}
				}

				if (ship.order instanceof HoldAndColoniseShipOrder || ship.order instanceof JumpToAndColonisePlanetShipOrder) {
					const targetHex = player.map.hex(ship.location);
					if (targetHex.planet && targetHex.owner) {
						ship.report = `${ship.location} is a class ${targetHex.level} colony belonging to ${Game.FULLNAMES[targetHex.owner]}, Unable to colonise`;
					} else if (targetHex.planet) {
						ship.report =`colonisation of ${ship.location} has been achieved`;
						Game.State.baseMap.colonise(ship.location,player.stylename,ship.getColonyImprovement());
						shipsToBeDestroyed.push(ship);
					}
					if (ship.hasAReport()) {
						report = ship.report;
					}
				}

				if (ship.order instanceof HoldPositionShipOrder) {
					player.map.setLastScanned(ship.location,Game.State.turn);
					const targetHex = player.map.hex(ship.location);
					if (targetHex.planet && targetHex.owner) {
						ship.report = `${ship.location} is a class ${targetHex.level} colony belonging to ${Game.FULLNAMES[targetHex.owner]}`;
					} else if (targetHex.planet) {
						ship.report = `${ship.location} is an uncolonised world`;
					}
				}

				if (report) updates.push(ship.name+" reports that "+report);
			}

			// Now remove the ships no longer needed
			player.removeShips(shipsToBeDestroyed);
		}

		// Expenditure orders
		for(let order of playerOrders) {
			let update = order.apply(Game._players[0],Game.State.baseMap);
			if (update) {
				updates.push(update);
			}
		}
		// Scanning is done when the scanner table is rebuilt
		if (updates.length > 0) {
			Game._showMap(Game._players[0].map); // To show the ship discoveries
			summary.classList.add("hidden");
			proc.classList.remove("hidden");
			const list = document.getElementById("messages");
			list.replaceChildren();
			for(let update of updates) {
				const element = document.createElement("li");
				element.append(update);
				list.append(element);
			}
		} else {
			Game._newTurn();
		}
	},
	_completeTurn: function() {
		let summary = document.getElementById("summary");
		let proc = document.getElementById("processOrder");
		proc.classList.add("hidden");
		summary.classList.remove("hidden");
		Game._newTurn();
	},
	_showLoadButton: function() {
		if (localStorage.domainOfTheAncientsSaveDate) {
			const saved = new Date(localStorage.domainOfTheAncientsSaveDate);
			const nextUpdate = Util.next_time_diff(saved);
			if (localStorage.domainOfTheAncientsTurnNumber) Util.replaceText("savegameturn",localStorage.domainOfTheAncientsTurnNumber);
			Util.replaceText("savegamedate",Util.time_diff(saved));
			// Is the welcome screen hidden?
			const hiddenWelcome = document.getElementById("welcome").classList.contains("hidden");
			if (nextUpdate && !hiddenWelcome) {
				window.setTimeout(Util.captureExceptionFunction(Game._showLoadButton, nextUpdate*1000));
			}
		}
	},
	_saveGameState: function() {
		let data = {};
		data.ROTRNGstate = JSON.stringify(ROT.RNG.getState());
		data.turn = Game.State.turn;
		data.baseMap = Game.State.baseMap;
		data.playerTech = Game._players[0].techIndex;
		data.players = [];
		for(let p of Game._players) {
			data.players.push(p);
		}
		data.built = [];
		for(let design of Game._standardShipDesigns) {
			data.built.push(design._built);
		}
		localStorage.domainOfTheAncientsTurnNumber = Game.State.turn;
		localStorage.domainOfTheAncientsSaveDate = (new Date()).toISOString();
		let version = {game: Game.State.version, plugin:"3.173.0"};
		localStorage.domainOfTheAncientsVersion = JSON.stringify(version);
		localStorage.domainOfTheAncients = JSON.stringify(data);
	},
	_load: function() {
		const data = JSON.parse(localStorage.domainOfTheAncients);
		ROT.RNG.setState(JSON.parse(data.ROTRNGstate));
		Game.State.turn = data.turn;
		Game.State.baseMap = new Map(data.baseMap);
		Game._createPlayers(data.playerTech);
		for(let i=0; i<Game._players.length; i++) {
			Game._players[i].fromJSON(data.players[i],Game.State.baseMap);
		}

		let i=0;
		for(let design of Game._standardShipDesigns) {
			design._built = data.built[i++];
		}

		PlayerOrders.setPlayer(Game._players[0]);
		PlayerOrders.setStandardShipDesigns(Game._standardShipDesigns);

		const v = JSON.parse(localStorage.domainOfTheAncientsVersion);
		Game.State.version = v.game;

		document.getElementById("welcome").classList.replace("visible", "hidden");
		document.getElementById("summary").classList.replace("hidden", "visible");
		Game._startTurn();
	},
};

window.onload = function() {
	Util.captureException(Game.init);
};
