/*!
 * © 2022 David Vines
 *
 * domainOfTheAncients is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * any later version.
 *
 * domainOfTheAncients is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with domainOfTheAncients. If not, see https://www.gnu.org/licenses/gpl-2.0.html.
 */
/* globals Game, ShipDesignType */
/* exported ResearchOrder, CarryForwardOrder, ImproveColonyOrder, AddMarinesOrder, AddSensorBuildingOrder, BuyVehicleOrder, BuyStarBaseOrder */
class ResearchOrder {
	constructor(tech,amount) {
		this._tech = tech;
		this._amount = amount;
	}
	toString() {
		return "Research "+this._amount+" in "+this._tech.option;
	}
	apply(player,_map) {
		if (player.addResearch(this._tech,this._amount))
		{
			const level = player.researchLevel(this._tech);
			const item = document.getElementById(this._tech.domPrefix+level);
			return this._tech.option+" technology has now reached level "+level+". "+item.innerText+"."+(item.title ? " "+item.title+"." : "");
		}
		return null;
	}
}

class CarryForwardOrder {
	constructor(amount) {
		this._amount = amount;
	}
	toString() {
		return "carry forward "+this._amount+" PPs";
	}
	apply(player,_map) {
		player.carriedProduction = this._amount;
		return null;
	}
}

class ImproveColonyOrder {
	constructor(colony) {
		this._colony = colony;
	}
	get colony() {
		return this._colony;
	}
	toString() {
		return "Improve colony at "+this._colony;
	}
	apply(_player,map) {
		map.improveColony(this._colony);
		return null;
	}
}

class AddMarinesOrder {
	constructor(colony,companies) {
		this._colony = colony;
		this._companies = companies;
	}
	get colony() {
		return this._colony;
	}
	get companies() {
		return this._companies;
	}
	toString() {
		return "Add "+this._companies+" marine companies to the  colony at "+this._colony;
	}
	apply(_player,map) {
		map.addMarines(this._colony,this._companies);
		return null;
	}
}

class AddSensorBuildingOrder {
	constructor(colony) {
		this._colony = colony;
	}
	get colony() {
		return this._colony;
	}
	toString() {
		return "Add a sensor building to the colony at "+this._colony;
	}
	apply(_player,map) {
		map.addSensorBuilding(this._colony);
		return null;
	}
}

class BuyVehicleOrder {
	constructor(design,name,location) {
		this._shipDesign = design;
		this._name = name;
		this._location = location;
	}
	get design() {
		return this._shipDesign;
	}
	get startLocation() {
		return this._location;
	}
	get name() {
		return this._name;
	}
	toString() {
		return `Add a new ${this._design.name} called '${this._name}'+at ${this._location}`;
	}
	apply(player,map) {
		player.addShip(this._shipDesign.build(player,this._name,this._location));
		if (this._shipDesign.type === ShipDesignType.StarBase) {
			map.addStarBase(this._location,player.stylename);
		}
		return null;
	}
}

class BuyStarBaseOrder {
	constructor(design,name,location) {
		this._shipDesign = design;
		this._name = name;
		this._location = location;
	}
	get design() {
		return this._shipDesign;
	}
	get startLocation() {
		return this._location;
	}
	get name() {
		return this._name;
	}
	toString() {
		return `Add a new ${this._design.name} called '${this._name}'+at ${this._location}`;
	}
	apply(player,map) {
		// The star base can only be deployed to empty hexes - the star base is discarded if the hex has anything there
		const hex = map.hex(this._location);
		if (hex.owner && hex.owner !== player.stylename) {
			if (hex.level > 0) {
				return `Cannot deploy ${this._name} to ${this._location}. It is already occupied by a class ${hex.level} colony. Our Star Base has been destroyed to prevent its capture by ${Game.FULLNAMES[hex.owner]}`;
			} else {
				return `Cannot deploy ${this._name} to ${this._location}. It is already occupied by another Star Base. Our Star Base has been destroyed to prevent its capture by ${Game.FULLNAMES[hex.owner]}`;
			}
		} else {
			player.addShip(this._shipDesign.build(player,this._name,this._location));
			if (this._shipDesign.type === ShipDesignType.StarBase) {
				map.addStarBase(this._location,player.stylename);
			}
			return null;
		}
	}
}
