/*!
 * © 2022 David Vines
 *
 * domainOfTheAncients is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 2 of the License, or
 * any later version.
 *
 * domainOfTheAncients is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with domainOfTheAncients. If not, see https://www.gnu.org/licenses/gpl-2.0.html.
 */
/* globals ALLTECHS, Map, ShipDesign, Ship */
/* exported Player */
class Player {
	constructor(stylename,capital,indexOfFavouredTech) {
		let techselect = document.getElementById("favouredTech");
		this._stylename = stylename;
		this._indexOfFavouredTech = indexOfFavouredTech;
		this._tech = techselect.options[indexOfFavouredTech].text;
		this._carriedProduction = 0;
		this._research = Array.from(ALLTECHS, _tech => { return 0; });
		this._production = 400;
		this._shipDesigns = [];
		this._ships = [];
		this._hasBuiltAShip = false;
		this._capital = capital;
		this._map = null;
	}
	init(baseMap,turn) {
		this._map = new Map(baseMap);
		const stylename = this._stylename;
		const map = this._map;
		map.forEachHex(function(hex) {
			if (hex.owner == stylename) {
				map.setLastScanned(hex.id,turn);
			}
		});

	}
	get map() {
		return this._map;
	}
	computeCommsGrid() {
		this._map.computeCommsGrid(this);
	}
	get stylename() {
		return this._stylename;
	}
	get capital() {
		return this._capital;
	}
	get production() {
		return this._production;
	}
	set production(p) {
		this._production = p;
	}
	get carriedProduction() {
		return this._carriedProduction;
	}
	set carriedProduction(cp) {
		this._carriedProduction = cp;
	}
	get techIndex() {
		return this._indexOfFavouredTech;
	}
	get tech() {
		return this._tech;
	}
	get shipDesigns() {
		return this._shipDesigns;
	}
	get hasBuiltAShip() {
		return this._hasBuiltAShip;
	}
	addShipDesign(design) {
		this._shipDesigns.push(design);
	}
	addShip(ship) {
		this._ships.push(ship);
		this._hasBuiltAShip = true;
	}
	removeShips(destroyedShips) {
		this._ships = this._ships.filter(x => !destroyedShips.includes(x));
	}
	get ships() {
		return this._ships;
	}
	researchLevel(tech) {
		return ALLTECHS[tech.index].level(this._research[tech.index]);
	}
	addResearch(tech,points) {
		let oldLevel = this.researchLevel(tech);
		this._research[tech.index] += parseInt(points);
		let newLevel = this.researchLevel(tech);
		return newLevel > oldLevel;
	}
	forEachTech(f) {
		let self = this;
		ALLTECHS.forEach(function(tech,index) {
			f(tech,self._research[index]);
		});
	}
	toJSON() {
		let data = {carried: this._carriedProduction,
			capital: this._capital,
			research: this._research,
			production: this._production,
			builtAShip: this._hasBuiltAShip,
			map: this._map,
			shipDesigns: [],
			ships: []
		};
		for(let design of this._shipDesigns) {
			data.shipDesigns.push(design);
		}
		for(let ship of this._ships) {
			data.ships.push(ship);
		}
		return data;
	}
	fromJSON(data,baseMap) {
		this._capital = data.capital;
		this._production = data.production;
		this._carriedProduction = data.carried;
		this._research = data.research;
		this._hasBuiltAShip = data.builtAShip;
		if (data.map) {
			this._map = new Map(data.map,baseMap);
		}
		for(let design of data.shipDesigns) {
			this._shipDesigns.push(ShipDesign.fromJSON(design));
		}
		for(let ship of data.ships) {
			this._ships.push(Ship.fromJSON(ship));
		}
	}
}